/** This module handles stored options, currently all except input maps.
*
* The purpose of having all options centrally controlled is to allow generic handling by the GUI
* and ease saving and loading of values.
*/
module mde.options;

import mde.exception;

import mde.mergetag.Reader;
import mde.mergetag.Writer;
import mde.mergetag.DataSet;
import mde.mergetag.exception;
import mde.resource.paths;

import tango.scrapple.text.convert.parseTo : parseTo;
import tango.scrapple.text.convert.parseFrom : parseFrom;

import tango.util.log.Log : Log, Logger;

Options options;
static this() {
    // Since options is used in many places, creating a class of default values to use until
    // options have been loaded makes its use safer.
    // In addition, loading should be atomic.
    options = new Options();
}

// Later: add automatic saving/loading for variables, variable lists with GUI name and description.

/** Class instance stores all options.
*
* All options and handling should be non-static to allow possibility of profiles later.
*/
class Options : IDataSection
{
    /* The options.
    *
    * For now, named variables. Later, lists (and named variables?).
    *
    * They can be read and set directly by other code (so not thread-safe for writing currently).
    */
    char[] L10n;        // locale, e.g. en-GB
    bool useThreads;    // set 0 to disable threading
    int logLevel;       // tango logger level
    
    /* Have any options been changed? Only bother writing if true.
    *
    * Maybe update this in the future to save if the last file edit was not made by mde. */
    bool changed = false;
    
    /* The code to load/save the values.
    *
    * Uses mergetag.
    */
    void addTag (char[] tp, ID id, char[] dt) {
        if (tp == "char[]") {
            if (id == cast(ID)"L10n") L10n = parseTo!(char[]) (dt);
        } else if (tp == "bool") {
            if (id == cast(ID)"useThreads") useThreads = parseTo!(bool) (dt);
        } else if (tp == "int") {
            if (id == cast(ID)"logLevel") logLevel = parseTo!(int) (dt);
        }
    }
    void writeAll (ItemDelg dlg) {
        dlg ("char[]", cast(ID)"L10n", parseFrom!(char[]) (L10n));
        dlg ("bool", cast(ID)"useThreads", parseFrom!(bool) (useThreads));
        dlg ("int", cast(ID)"logLevel", parseFrom!(int) (logLevel));
    }
    
    
    /* Load/save options from file.
    *
    * If the file doesn't exist, no reading is attempted (options are left at default values).
    */
    private static const fileName = "options";
    private static const secName = cast(ID)"Default";
    static void load () {
        // Check it exists (if not it should still be created on exit).
        // Don't bother checking it's not a folder, because it could still be a block or something.
        if (!confDir.exists (fileName)) return;
        
        IReader reader;
        try {
            reader = confDir.makeMTReader (fileName, PRIORITY.LOW_HIGH);
            reader.dataSecCreator = delegate IDataSection(ID) {
                return new Options;
            };
            reader.read;
        } catch (MTException e) {
            logger.error ("Mergetag exception occurred:");
            logger.error (e.msg);
            throw new optionsLoadException ("Loading aborted: mergetag exception");
        }
        
        IDataSection* secP = secName in reader.dataset.sec;
        Options o = cast(Options) *secP;
        if (o is null) {
            throw new optionsLoadException ("Loading failed: expected section not found");
        }
        else options = o;   // loading was succesful.
    }
    static void save () {
        if (!options.changed) return;   // skip

        DataSet ds = new DataSet();
        ds.sec[secName] = options;
        
        IWriter writer;
        try {
            writer = confDir.makeMTWriter (fileName, ds);
            writer.write();
        } catch (MTException e) {
            logger.error ("Mergetag exception occurred; saving aborted:");
            logger.error (e.msg);
            //FIXME: currently nothing done besides logging the error
        }
    }
    
    private static Logger logger;
    static this() {
        logger = Log.getLogger ("mde.options");
    }
}
